########## IMPORTS ##########
from CAA_class import _CAA
from OAA_class import _OAA
from RBOAA_class import _RBOAA
from TSAA_class import _TSAA
from synthetic_data_class import _synthetic_data
import pandas as pd
import numpy as np
import pickle
from os import path


########## ARCHETYPAL ANALYSIS MODULE CLASS ##########
class AA:
    
    def __init__(self):
        
        self._CAA = _CAA()
        self._OAA = _OAA()
        self._RBOAA = _RBOAA()
        self._TSAA = _TSAA()
        self._results = {"CAA": [], "OAA": [], "RBOAA": [], "TSAA": []}
        self._synthetic_results = {"CAA": [], "OAA": [], "RBOAA": [], "TSAA": []}
        self._has_data = False
        self.has_synthetic_data = False
        self.has_dataframe = False
        self.has_archetype_dataframe = False
        self.archetype_dataframe = pd.DataFrame()
        self.has_ranked_archetype_dataframe = False
        self.ranked_archetype_dataframe = pd.DataFrame()


    def load_data(self, X: np.ndarray, columns: list()):
        self.columns = columns
        self.X = X
        self.N, self.M = X.shape
        self._has_data = True
        if self.N<self.M:
            print("Your data has more attributes than subjects.")
            print(f"Your data has {self.M} attributes and {self.N} subjects.")
            print("This is highly unusual for this type of data.")
            print("Please try loading transposed data instead.")
        else:
            print(f"\nThe data was loaded successfully!\n")


    def load_csv(self, filename: str, columns: list(), rows: int = None, mute: bool = False):
        self.columns, self.M, self.N, self.X = self._clean_data(filename, columns, rows)
        self._has_data = True
        if not mute:
            print(f"\nThe data of \'{filename}\' was loaded successfully!\n")

    
    def _clean_data(self, filename, columns, rows):
        df = pd.read_csv(filename)

        column_names = df.columns.to_numpy()
        
        if not columns is None:
            column_names = column_names[columns]
            X = df[column_names]
        else:
            X = df[column_names]
        if not rows is None:
            X = X.iloc[range(rows),:]
            self.has_dataframe = True
            self.dataframe = df.iloc[range(rows),:]
        else:
            self.has_dataframe = True
            self.dataframe = df

        X = X.to_numpy().T
        M, N = X.shape

        return column_names, M, N, X
    

    def create_synthetic_data(self, N: int = 1000, M: int = 10, K: int = 3, p: int = 6, sigma: float = -20.0, rb: bool = False, b_param: float = 100, a_param: float = 1, sigma_dev: float = 0, mute = False):
        if N < 2:
            print("The value of N can't be less than 2. The value specified was {0}".format(N))
        elif M < 2:
            print("The value of M can't be less than 2. The value specified was {0}".format(M))
        elif K < 2:
            print("The value of K can't be less than 2. The value specified was {0}".format(K))
        elif p < 2:
            print("The value of p can't be less than 2. The value specified was {0}".format(p))
        else:
            self._synthetic_data = _synthetic_data(N, M, K, p, sigma, rb, a_param, b_param, sigma_std = sigma_dev)
            self.has_synthetic_data = True
            self._synthetic_results = {"CAA": [], "OAA": [], "RBOAA": [], "TSAA": []}
            if not mute:
                print("\nThe synthetic data was successfully created! To use the data in an analysis, specificy the with_synthetic_data parameter as True.\n")


    def analyse(self, K: int = 3, p: int = 6, n_iter: int = 1000, early_stopping: bool = True, model_type = "all", lr: float = 0.01, mute: bool = False, with_synthetic_data: bool = False, with_hot_start: bool = False):
        
        success = True

        if model_type == "TSAA":
            print("The model_type TSAA has been deprecated, due to errors in the method.")
            success = False

        elif self._has_data and not with_synthetic_data:
            if model_type == "all" or model_type == "CAA":
                self._results["CAA"].insert(0,self._CAA._compute_archetypes(self.X, K, p, n_iter, lr, mute,self.columns,early_stopping=early_stopping))
            elif model_type == "all" or model_type == "OAA":
                self._results["OAA"].insert(0,self._OAA._compute_archetypes(self.X, K, p, n_iter, lr, mute,self.columns,with_synthetic_data=False,early_stopping=early_stopping))
            elif model_type == "all" or model_type == "RBOAA":
                self._results["RBOAA"].insert(0,self._RBOAA._compute_archetypes(self.X, K, p, n_iter, lr, mute,self.columns, with_synthetic_data=False, early_stopping=early_stopping, with_OAA_initialization = with_hot_start))
            elif model_type == "all" or model_type == "TSAA":
                self._results["TSAA"].insert(0,self._TSAA._compute_archetypes(self.X, K, p, n_iter, lr, mute,self.columns,early_stopping=early_stopping))
            else:
                print("The model_type \"{0}\" specified, does not match any of the possible AA_types.".format(model_type))
                success = False
    
        elif self.has_synthetic_data and with_synthetic_data:
            if model_type == "all" or model_type == "CAA":
                self._synthetic_results ["CAA"].insert(0,self._CAA._compute_archetypes(self._synthetic_data.X, K, p, n_iter, lr, mute, self._synthetic_data.columns, with_synthetic_data=True,early_stopping=early_stopping))
            elif model_type == "all" or model_type == "OAA":
                self._synthetic_results["OAA"].insert(0,self._OAA._compute_archetypes(self._synthetic_data.X, K, p, n_iter, lr, mute, self._synthetic_data.columns,with_synthetic_data=True,early_stopping=early_stopping,for_hotstart_usage=False))
            elif model_type == "all" or model_type == "RBOAA":
                self._synthetic_results["RBOAA"].insert(0,self._RBOAA._compute_archetypes(self._synthetic_data.X, K, p, n_iter, lr, mute, self._synthetic_data.columns,with_synthetic_data=True,early_stopping=early_stopping,with_OAA_initialization = with_hot_start))
            elif model_type == "all" or model_type == "TSAA":
                self._synthetic_results["TSAA"].insert(0,self._TSAA._compute_archetypes(self._synthetic_data.X, K, p, n_iter, lr, mute, self._synthetic_data.columns,with_synthetic_data=True,early_stopping=early_stopping))
            else:
                print("The model_type \"{0}\" specified, does not match any of the possible AA_types.".format(model_type))
                success = False
        
        else:
            print("\nYou have not loaded any data yet! \nPlease load data through the \'load_data\' or \'load_csv\' methods and try again.\n")
            success = False

        if success and self.has_dataframe:
            self.create_dataframe(model_type=model_type,with_synthetic_data=with_synthetic_data,mute=True)
            self.create_dataframe(model_type=model_type,with_synthetic_data=with_synthetic_data,archetype_rank=3,mute=True)


    def plot(self, 
            model_type: str = "CAA", 
            plot_type: str = "PCA_scatter_plot", 
            title: str = "",
            save_figure: bool = False,
            filename: str = "figure",
            result_number: int = 0, 
            attributes: list() = [1,2], 
            archetype_number: int = 1, 
            types: dict = {"type 1": [1],"type 2": [2]},
            weighted: str = "equal_norm",
            subject_indexes: list() = [1],
            attribute_indexes: list() = [],
            with_synthetic_data: bool = False):
        
        if not model_type in ["CAA", "OAA", "RBOAA", "TSAA"]:
            print("\nThe model type you have specified can not be recognized. Please try again.")
        elif not plot_type in ["PCA_scatter_plot","attribute_scatter_plot","loss_plot","mixture_plot","barplot","barplot_all","typal_plot","pie_chart","attribute_distribution","circular_typal_barplot"]:
            print("\nThe plot type you have specified can not be recognized. Please try again.\n")
        elif not weighted in ["none","equal_norm","equal","norm"]:
            print(f"\nThe \'weighted\' parameter received an unexpected value of {weighted}.\n")
        elif not attribute_indexes == [] and not self.has_archetype_dataframe:
            print(f"\nYou have not created any dataframe to plot w.r.t..\n")
        
        elif not with_synthetic_data:
            if result_number < 0 or not result_number < len(self._results[model_type]):
                print("\nThe result you are requesting to plot is not availabe.\n Please make sure you have specified the input correctly.\n")
            elif archetype_number < 1 or archetype_number > self._results[model_type][result_number].K:
                print(f"\nThe \'archetype_number\' parameter received an unexpected value of {archetype_number}.\n")
            elif any(np.array(attributes) < 0) or any(np.array(attributes) > len(self._results[model_type][result_number].columns)):
                print(f"\nThe \'attributes\' parameter received an unexpected value of {attributes}.\n")
            elif any(np.array(subject_indexes) < 0) or any(np.array(subject_indexes) > self._results[model_type][result_number].N-1):
                print(f"\nThe \'subject_indexes\' parameter received an unexpected value of {subject_indexes}.\n")
            else:
                result = self._results[model_type][result_number]
                result._plot(plot_type,attributes,archetype_number,types,weighted,subject_indexes,attribute_indexes, self.archetype_dataframe ,save_figure,filename,title)
                if save_figure:
                    print("\nThe requested plot was successfully saved to your device!\n")
                else:
                    print("\nThe requested plot was successfully plotted!\n")
        
        else:
            if result_number < 0 or not result_number < len(self._synthetic_results[model_type]):
                print("\nThe result you are requesting to plot is not available.\n Please make sure you have specified the input correctly.\n")
            elif archetype_number < 0 or archetype_number > self._synthetic_results[model_type][result_number].K:
                print(f"\nThe \'archetype_number\' parameter received an unexpected value of {archetype_number}.\n")
            elif any(np.array(attributes) < 0) or any(np.array(attributes) > len(self._synthetic_results[model_type][result_number].columns)):
                print(f"\nThe \'attributes\' parameter received an unexpected value of {attributes}.\n")
            elif any(np.array(subject_indexes) < 0) or any(np.array(subject_indexes) > self._synthetic_results[model_type][result_number].N-1):
                print(f"\nThe \'subject_indexes\' parameter received an unexpected value of {subject_indexes}.\n")
            else:
                result = self._synthetic_results[model_type][result_number]
                result._plot(plot_type,attributes,archetype_number,types,weighted,subject_indexes,attribute_indexes, self.archetype_dataframe,save_figure,filename,title)
                print("\nThe requested synthetic data result plot was successfully plotted!\n")


    def save_analysis(self,filename: str = "analysis",model_type: str = "CAA", result_number: int = 0, with_synthetic_data: bool = False, save_synthetic_data: bool = True):

        if not model_type in ["CAA", "OAA", "RBOAA", "TSAA"]:
            print("\nThe model type you have specified can not be recognized. Please try again.\n")
        
        if not with_synthetic_data:
            if not result_number < len(self._results[model_type]):
                print("\nThe analysis you are requesting to save is not available.\n Please make sure you have specified the input correctly.\n")
            
            
            else:
                self._results[model_type][result_number]._save(filename)
                print("\nThe analysis was successfully saved!\n")
        else:
            if not result_number < len(self._synthetic_results[model_type]):
                print("\nThe analysis with synthetic data, which you are requesting to save is not available.\n Please make sure you have specified the input correctly.\n")
            
            else:
                self._synthetic_results[model_type][result_number]._save(filename)
                if save_synthetic_data:
                    self._synthetic_data._save(model_type,filename)
                print("\nThe analysis was successfully saved!\n")

    
    def load_analysis(self, filename: str = "analysis", model_type: str = "CAA", with_synthetic_data: bool = False):
        if not model_type in ["CAA", "OAA", "RBOAA", "TSAA"]:
            print("\nThe model type you have specified can not be recognized. Please try again.\n")

        elif not with_synthetic_data:
            if not path.exists("results/" + model_type + "_" + filename + '.obj'):
                print(f"The analysis {filename} of type {model_type} does not exist on your device.")
            
            
            else:
                file = open("results/" + model_type + "_" + filename + '.obj','rb')
                result = pickle.load(file)
                file.close()
                self._results[model_type].append(result)
                print("\nThe analysis was successfully loaded!\n")

        else:
            if not path.exists("synthetic_results/" + model_type + "_" + filename + '.obj'):
                print(f"The analysis {filename} with synthetic data of type {model_type} does not exist on your device.")
            
            else:
                file = open("synthetic_results/" + model_type + "_" + filename + '.obj','rb')
                result = pickle.load(file)
                file.close()
                self._synthetic_results[model_type].append(result)

                file = open("synthetic_results/" + model_type + "_" + filename + '_metadata' + '.obj','rb')
                result = pickle.load(file)
                file.close()
                self._synthetic_data = result

                print("\nThe analysis with synthetic data was successfully loaded!\n")
                self.has_synthetic_data = True


    def create_dataframe(self, model_type: str = "CAA", result_number: int = 0, with_synthetic_data: bool = False, archetype_rank: int = 0, return_dataframe: bool = False, mute = False):
        
        if not model_type in ["CAA", "OAA", "RBOAA", "TSAA"]:
            print("\nThe model type you have specified can not be recognized. Please try again.\n")
            return

        if (with_synthetic_data and len(self._synthetic_results[model_type]) <= result_number) or (not with_synthetic_data and len(self._results[model_type]) <= result_number):
            print("\nThe result you have specified to create the dataframe can not be recognized. Please try again.\n")
            return

        if with_synthetic_data:
            result = self._synthetic_results[model_type][result_number]
        else:
            result = self._results[model_type][result_number]

        if archetype_rank == False:
            if self.has_dataframe:
                self.archetype_dataframe = self.dataframe.copy()
                for archetype in range(result.K):
                    self.archetype_dataframe["Archetype {0}".format(archetype+1)] = result.A[archetype,:]
                if not mute:
                    print("\nThe dataframe was successfully created from a copy of your imported dataframe.\n")
                
            else:
                dict = {}
                for archetype in range(result.K):
                    dict["Archetype {0}".format(archetype+1)] = result.A[archetype,:]
                archetype_dataframe = pd.DataFrame.from_dict(dict)
                self.archetype_dataframe = archetype_dataframe
                if not mute:
                    print("\nThe dataframe was successfully created.\n")

        else:
            if self.has_dataframe:
                self.ranked_archetype_dataframe = self.dataframe.copy()
                for rank in range(archetype_rank):
                    rank_list = []
                    for n in range(result.N):
                        rank_list.append(np.where(result.A[:,n] == np.sort(result.A[:,n])[::-1][rank])[0][0]+1)
                    self.ranked_archetype_dataframe["Archetype Rank {0}".format(rank+1)] = rank_list
                if not mute:
                    print("\nThe dataframe was successfully created from a copy of your imported dataframe.\n")

        if archetype_rank == False:
            self.has_archetype_dataframe = True
            if return_dataframe:
                return self.archetype_dataframe
        else:
            self.has_ranked_archetype_dataframe = True
            if return_dataframe:
                return self.ranked_archetype_dataframe


    def get_dataframe(self,ranked_dataframe: bool = False):

        if not ranked_dataframe:
            if not self.has_archetype_dataframe:
                print("\nThe dataframe which you have requested, does not exist yet.\n")
            
            return self.archetype_dataframe
        
        else:
            if not self.has_ranked_archetype_dataframe:
                print("\nThe dataframe which you have requested, does not exist yet.\n")
            
            return self.ranked_archetype_dataframe


    def get_analysis(self, model_type: str = "CAA", result_number: int = 0, with_synthetic_data: bool = False):

        if not model_type in ["CAA", "OAA", "RBOAA", "TSAA"]:
            print("\nThe model type you have specified can not be recognized. Please try again.\n")
            return

        if (with_synthetic_data and len(self._synthetic_results[model_type]) <= result_number) or (not with_synthetic_data and len(self._results[model_type]) <= result_number):
            print("\nThe result you have specified to create the dataframe can not be recognized. Please try again.\n")
            return

        if with_synthetic_data:
            result = self._synthetic_results[model_type][result_number]
        else:
            result = self._results[model_type][result_number]
        
        return result



########## IMPORTS ##########
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from timeit import default_timer as timer
from AA_result_class import _OAA_result
from loading_bar_class import _loading_bar
from CAA_class import _CAA



########## ORDINAL ARCHETYPAL ANALYSIS CLASS ##########
class _OAA:

    ########## HELPER FUNCTION // EARLY STOPPING ##########
    def _early_stopping(self):
        next_imp = self.loss[-round(len(self.loss)/100)]-self.loss[-1]
        prev_imp = (self.loss[0]-self.loss[-1])*1e-5
        return next_imp < prev_imp

    ########## HELPER FUNCTION // A AND B ##########
    def _apply_constraints_AB(self,A):
        m = nn.Softmax(dim=1)
        return m(A)

    ########## HELPER FUNCTION // BETAS ##########
    def _apply_constraints_beta(self,b): 
        m = nn.Softmax(dim=0)
        return torch.cumsum(m(b), dim=0)[:len(b)-1]

    ########## HELPER FUNCTION // SIGMA ##########
    def _apply_constraints_sigma(self,sigma):
        m = nn.Softplus()
        return m(sigma)

    ########## HELPER FUNCTION // ALPHA ##########
    def _calculate_alpha(self,b):
        b_j = torch.cat((torch.tensor([0.0]),b),0)
        b_j_plus1 = torch.cat((b,torch.tensor([1.0])),0)
        alphas = (b_j_plus1+b_j)/2
        return alphas

    ########## HELPER FUNCTION // X_tilde ##########
    def _calculate_X_tilde(self,X,alphas):
        X_tilde = alphas[X.long()-1]
        return X_tilde

    ########## HELPER FUNCTION // X_hat ##########
    def _calculate_X_hat(self,X_tilde,A,B):
        Z = B @ X_tilde
        X_hat = A @ Z
        return X_hat
    
    ########## HELPER FUNCTION // LOSS ##########
    def _calculate_loss(self,Xt, X_hat, b, sigma):
        
        pad = nn.ConstantPad1d(1, 0)
        b = pad(b)
        b[-1] = 1.0

        z_next = (b[Xt] - X_hat)/sigma
        z_prev = (b[Xt-1] - X_hat)/sigma
        z_next[Xt == len(b)+1] = np.inf
        z_prev[Xt == 1] = -np.inf

        P_next = torch.distributions.normal.Normal(0, 1).cdf(z_next)
        P_prev = torch.distributions.normal.Normal(0, 1).cdf(z_prev)
        neg_logP = -torch.log(( P_next - P_prev ) +1e-10)
        loss = torch.sum(neg_logP)

        return loss

    ########## HELPER FUNCTION // ERROR ##########
    def _error(self,Xt,A_non_constraint,B_non_constraint,b_non_constraint,sigma_non_constraint):

        A = self._apply_constraints_AB(A_non_constraint)
        B = self._apply_constraints_AB(B_non_constraint)
        b = self._apply_constraints_beta(b_non_constraint)
        sigma = self._apply_constraints_sigma(sigma_non_constraint)
        alphas = self._calculate_alpha(b)
        
        X_tilde = self._calculate_X_tilde(Xt,alphas)
        X_hat = self._calculate_X_hat(X_tilde,A,B)

        loss = self._calculate_loss(Xt, X_hat, b, sigma)
        
        return loss
        

    ########## COMPUTE ARCHETYPES FUNCTION OF OAA ##########
    def _compute_archetypes(
        self, 
        X, 
        K, 
        p, 
        n_iter, 
        lr, 
        mute, 
        columns, 
        with_synthetic_data = False, 
        early_stopping = False, 
        for_hotstart_usage = False):


        ########## INITIALIZATION ##########
        self.N, self.M = len(X.T), len(X.T[0,:])
        Xt = torch.tensor(X.T, dtype = torch.long)
        self.loss = []
        start = timer()

        A_non_constraint = torch.autograd.Variable(torch.randn(self.N, K), requires_grad=True)
        B_non_constraint = torch.autograd.Variable(torch.randn(K, self.N), requires_grad=True)
        b_non_constraint = torch.autograd.Variable(torch.rand(p), requires_grad=True)
        sigma_non_constraint = torch.autograd.Variable(torch.rand(1), requires_grad=True)

        optimizer = optim.Adam([A_non_constraint, 
                                B_non_constraint, 
                                b_non_constraint, 
                                sigma_non_constraint], amsgrad = True, lr = lr)

        if not mute:
            loading_bar = _loading_bar(n_iter, "Ordinal Archetypal Analysis")

        
        ########## ANALYSIS ##########
        for i in range(n_iter):
            if not mute:
                loading_bar._update()
            optimizer.zero_grad()
            L = self._error(Xt,A_non_constraint,B_non_constraint,b_non_constraint,sigma_non_constraint)
            self.loss.append(L.detach().numpy())
            L.backward()
            optimizer.step()


            ########## EARLY STOPPING ##########
            if i % 25 == 0 and early_stopping:
                if len(self.loss) > 200 and self._early_stopping():
                    if not mute:
                        loading_bar._kill()
                        print("Analysis ended due to early stopping.\n")
                    break
            
        
        ########## POST ANALYSIS ##########
        A_f = self._apply_constraints_AB(A_non_constraint).detach().numpy()
        B_f = self._apply_constraints_AB(B_non_constraint).detach().numpy()
        b_f = self._apply_constraints_beta(b_non_constraint)
        alphas_f = self._calculate_alpha(b_f)
        X_tilde_f = self._calculate_X_tilde(Xt,alphas_f).detach().numpy()
        Z_tilde_f = (self._apply_constraints_AB(B_non_constraint).detach().numpy() @ X_tilde_f)
        sigma_f = self._apply_constraints_sigma(sigma_non_constraint).detach().numpy()
        X_hat_f = self._calculate_X_hat(X_tilde_f,A_f,B_f)
        end = timer()
        time = round(end-start,2)
        Z_f = B_f @ X_tilde_f


        ########## CREATE RESULT INSTANCE ##########
        result = _OAA_result(
            A_f.T,
            B_f.T,
            X,
            n_iter,
            b_f.detach().numpy(),
            Z_f.T,
            X_tilde_f.T,
            Z_tilde_f.T,
            X_hat_f.T,
            self.loss,
            K,
            p,
            time,
            columns,
            "OAA",
            sigma_f,
            with_synthetic_data=with_synthetic_data)

        if not mute:
            result._print()
        

        ########## RETURN RESULT IF REGULAR, RETURN MATRICIES IF HOTSTART USAGE ##########
        if not for_hotstart_usage:
            return result
        else:
            A_non_constraint_np = A_non_constraint.detach().numpy()
            B_non_constraint_np = B_non_constraint.detach().numpy()
            sigma_non_constraint_np = sigma_non_constraint.detach().numpy()
            b_non_constraint_np = b_non_constraint.detach().numpy()
            return A_non_constraint_np, B_non_constraint_np, sigma_non_constraint_np, b_non_constraint_np



import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch import optim
from timeit import default_timer as timer
from scipy.special import softmax
from AA_result_class import _CAA_result
from loading_bar_class import _loading_bar

class _TSAA:
    
    def _logOdds(self, X):

        Ordinals = range(int(min(X.flatten())), int(max(X.flatten()+1)))
        probs = [(np.count_nonzero(X.flatten() == e))/len(X.flatten()) for e in Ordinals]
        baseline = max(probs)
        logvals = [np.log(probs[i]/baseline) for i in range(len(probs))]

        return logvals
    
    def _applySoftmax(self,M):
        return softmax(M)
    
    def _convertScores(self, X):
        
        Ordinals = range(int(min(X.flatten())), int(max(X.flatten()+1)))
        thetas = self._applySoftmax(self._logOdds(X))
        scores = [1+((k+1)-1)*thetas[k] for k in range(len(Ordinals))]
        
        return scores
        
    def _projectOrdinals(self, X):
        
        M, N = X.shape
        X_hat = np.empty((M, N))
        
        scores = self._convertScores(X)
        
        for i in range(M):
            for j in range(N):
                idx = X[i,j]-1
                X_hat[i,j] = scores[idx]
        return X_hat
    
    def _error(self, X,B,A):
        return torch.norm(X - X@B@A, p='fro')**2
    
    def _apply_constraints(self, A):
        m = nn.Softmax(dim=0)
        return m(A)
    
    ############# Two-step ordinal AA #############
    def _compute_archetypes(self, X, K, p, n_iter, lr, mute,columns, with_synthetic_data = False, early_stopping = False):
        
        ##### Project the data #####
        # Xt = torch.tensor(X, dtype = torch.long)
        
        X_hat = self._projectOrdinals(X)
        X_hat = torch.tensor(X_hat)
        
        ########## INITIALIZATION ##########
        self.RSS = []
        start = timer()
        if not mute:
            loading_bar = _loading_bar(n_iter, "Conventional Arhcetypal Analysis")
        N, _ = X.T.shape
        A = torch.autograd.Variable(torch.rand(K, N), requires_grad=True)
        B = torch.autograd.Variable(torch.rand(N, K), requires_grad=True)
        optimizer = optim.Adam([A, B], amsgrad = False, lr = lr)
        

        ########## ANALYSIS ##########
        for i in range(n_iter):
            if not mute:
                loading_bar._update()
            optimizer.zero_grad()
            L = self._error(X_hat, self._apply_constraints(B).double(), self._apply_constraints(A).double())
            self.RSS.append(L.detach().numpy())
            L.backward()
            optimizer.step()
            
            ########## EARLY STOPPING ##########
            if i % 25 == 0 and early_stopping:
                if len(self.RSS) > 200 and (self.RSS[-round(len(self.RSS)/100)]-self.RSS[-1]) < ((self.RSS[0]-self.RSS[-1])*1e-4):
                    if not mute:
                        loading_bar._kill()
                        print("Analysis ended due to early stopping.\n")
                    break
            

        ########## POST ANALYSIS ##########
        A_f = self._apply_constraints(A).detach().numpy()
        B_f = self._apply_constraints(B).detach().numpy()
        Z_f = X @ self._apply_constraints(B).detach().numpy()
        
        X_hat_f = X_hat.detach().numpy()
        end = timer()
        time = round(end-start,2)
        result = _CAA_result(A_f, B_f, X, X_hat_f, n_iter, self.RSS, Z_f, K, p, time,columns,"TSAA", with_synthetic_data = with_synthetic_data)
        if not mute:
            result._print()

        return result
    


########## IMPORTS ##########
import torch
import torch.nn as nn
import torch.optim as optim
from timeit import default_timer as timer 
from loading_bar_class import _loading_bar
from AA_result_class import _CAA_result


########## CONVENTIONAL ARCHETYPAL ANALYSIS CLASS ##########
class _CAA:

    ########## HELPER FUNCTION // EARLY STOPPING ##########
    def _early_stopping(self):
        next_imp = self.RSS[-round(len(self.RSS)/100)]-self.RSS[-1]
        prev_imp = (self.RSS[0]-self.RSS[-1])*1e-5
        return next_imp < prev_imp

    ########## HELPER FUNCTION // CALCULATE ERROR FOR EACH ITERATION ##########
    def _error(self, X,B,A):
        return torch.norm(X - X@B@A, p='fro')**2
    
    ########## HELPER FUNCTION // A CONSTRAINTS ##########
    def _apply_constraints(self, A):
        m = nn.Softmax(dim=0)
        return m(A)
    
    ########## COMPUTE ARCHETYPES FUNCTION OF CAA ##########
    def _compute_archetypes(self, X, K, p, n_iter, lr, mute,columns,with_synthetic_data = False, early_stopping = False, for_hotstart_usage = False):

        ########## INITIALIZATION ##########
        self.RSS = []
        start = timer()
        if not mute:
            loading_bar = _loading_bar(n_iter, "Conventional Arhcetypal Analysis")
        N, _ = X.T.shape
        Xt = torch.tensor(X,requires_grad=False).float()
        A = torch.autograd.Variable(torch.rand(K, N), requires_grad=True)
        B = torch.autograd.Variable(torch.rand(N, K), requires_grad=True)
        optimizer = optim.Adam([A, B], amsgrad = True, lr = lr)
        


        ########## ANALYSIS ##########
        for i in range(n_iter):
            if not mute:
                loading_bar._update()
            optimizer.zero_grad()
            L = self._error(Xt, self._apply_constraints(B), self._apply_constraints(A))
            self.RSS.append(L.detach().numpy())
            L.backward()
            optimizer.step()

            ########## EARLY STOPPING ##########
            if i % 25 == 0 and early_stopping:
                if len(self.RSS) > 200 and self._early_stopping():
                    if not mute:
                        loading_bar._kill()
                        print("Analysis ended due to early stopping.\n")
                    break
        
        

        ########## POST ANALYSIS ##########
        A_f = self._apply_constraints(A).detach().numpy()
        B_f = self._apply_constraints(B).detach().numpy()
        Z_f = (Xt@self._apply_constraints(B)).detach().numpy()
        X_hat_f = X@B_f@A_f
        end = timer()
        time = round(end-start,2)
        result = _CAA_result(A_f, B_f, X, X_hat_f, n_iter, self.RSS, Z_f, K, p, time,columns,"CAA",with_synthetic_data = with_synthetic_data)

        if not mute:
            result._print()

        if not for_hotstart_usage:
            return result
        else:
            return A.detach().numpy(), B.detach().numpy()


from setuptools import setup, find_packages


setup(
    name='AA-module',
    version='1.0.7',
    license='MIT',
    author="Andreas Bigom, Michael Harborg, Oliver Elmgren & Marcus Presutti",
    author_email='andreasbigom@gmail.com',
    long_description = "This module can be used to perform, evaluate and analyse Conventional Archetypal Analysis, Ordinal Archetypal Analysis and Response Bias Ordinal Archetypal Analysis. The module is developed by Andreas Bigom, Michael Harborg, Marcus Presutti and Oliver Elmgren in a collaboration between the Technical University of Denmark and Copenhagen Business School in order to enable students to analyse human questionnaire data in an effective and meaningful way.",
    description = "Module for performing various archetypal analyses.",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/anden1305/AAM-Module',
    keywords='Archetypal Analysis',
    python_requires = ">=3.10.2",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
          'pandas',
          'numpy',
          'matplotlib',
          'torch',
          'scipy',
          'sklearn',
    ],
)


########## IMPORT ##########
#from turtle import color
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
import numpy as np
from sklearn.decomposition import PCA
import pandas as pd


########## PLOTS CLASS ##########
class _plots:

    def _PCA_scatter_plot(self,Z,X,type,save_fig,filename, title):
        
        pca = PCA(n_components=2)
        pca.fit(Z.T)
        
        Z_pca = pca.transform(Z.T)
        X_pca = pca.transform(X.T)
        
        plt.rcParams["figure.figsize"] = (10,10)
        plt.scatter(X_pca[:,0], X_pca[:,1], c ="black", s = 1)
        for a in range(len(Z[0,:])):
            plt.scatter(Z_pca[a,0], Z_pca[a,1], marker ="^", s = 500, label="Archetype {0}".format(a+1))
        plt.xlabel("Principal Component 1", fontsize=15)
        plt.ylabel("Principal Component 2", fontsize=15)
        if title == "":
            plt.title(f"PCA Scatter Plot of {type}", fontsize = 20)
        else:
            plt.title(title, fontsize = 20)
        plt.legend(prop={'size': 15})

        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)


    def _attribute_scatter_plot(self,Z,X,attributes,type,p, save_fig, filename,title):
        
        plt.rcParams["figure.figsize"] = (10,10)
        plt.scatter(X[attributes[0]-1,:]*p, X[attributes[1]-1,:]*p, c ="black", s = 1)
        for a in range(len(Z[0,:])):
            plt.scatter(Z[attributes[0]-1,a]*p, Z[attributes[1]-1,a]*p, marker ="^", s = 500, label="Archetype {0}".format(a+1))
        plt.xlabel(f"Attribute {attributes[0]}", fontsize=15)
        plt.ylabel(f"Attribute {attributes[1]}", fontsize=15)
        plt.legend(prop={'size': 15})
        if title == "":
            plt.title(f"Attribute Scatter Plot of {type}", fontsize = 20)
        else:
            plt.title(title, fontsize = 20)

        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)


    def _loss_plot(self,loss,type, save_fig, filename,title):
        plt.plot(loss, c="#2c6c8c")
        plt.xlabel(f"Iteration of {type}")
        plt.ylabel(f"Loss of {type}")
        if title == "":
            plt.title(f"Loss w.r.t. Itteration of {type}")
        else:
            plt.title(title)
        
        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)


    def _mixture_plot(self,Z,A,type, save_fig, filename,title):

        plt.rcParams["figure.figsize"] = (10,10)
        fig = plt.figure()
        ax = fig.add_subplot(111)

        K = len(Z.T)
        corners = []
        for k in range(K):
            corners.append([np.cos(((2*np.pi)/K)*(k)), np.sin(((2*np.pi)/K)*(k))])
            plt.plot(
                np.cos(((2*np.pi)/K)*(k)), 
                np.sin(((2*np.pi)/K)*(k)), 
                marker="o", markersize=12, 
                markeredgecolor="black", 
                zorder=10,
                label = "Archetype {0}".format(k+1))

        points_x = []
        points_y = []
        for p in A.T:
            x = 0
            y = 0
            for k in range(K):
                x += p[k] * np.cos(((2*np.pi)/K)*(k))
                y += p[k] * np.sin(((2*np.pi)/K)*(k))
            points_x.append(x)
            points_y.append(y)
        
        p = Polygon(corners, closed=False,zorder=0)
        ax.add_patch(p)
        ax.set_xlim(-1.1,1.1)
        ax.set_ylim(-1.1,1.1)
        ax.set_aspect('equal')
        if title == "":
            plt.title(f"Mixture Plot of {type}", fontsize = 20)
        else:
            plt.title(title, fontsize = 20)
        plt.scatter(points_x, points_y, c ="black", s = 1, zorder=5)
        plt.legend()
        
        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)


    def _barplot(self,Z,columns, archetype_number,type,p, save_fig, filename,title):
        
        plt.rcParams["figure.figsize"] = (10,10)
        archetype = Z.T[archetype_number-1]
        if type in ["OAA","RBOAA"]:
            archetype *=p
        fig, ax = plt.subplots()
        ax.set_ylabel('Value')
        plt.xlabel('Attributes')
        if title == "":
            ax.set_title(f"Value-Distribution of Archeype {archetype_number}")
        else:
            ax.set_title(title)
        ax.bar(np.arange(len(archetype)),archetype)
        ax.set_xticks(np.arange(len(archetype)))
        ax.set_xticklabels(labels=columns)
        plt.ylim(0, p+0.5)
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
        fig.set_size_inches(10, 10)
        
        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)


    def _barplot_all(self,Z,columns,type, p,  save_fig, filename,title):
        plt.rcParams["figure.figsize"] = (10,10)
        data = []
        names = ["Attributes"]
        
        for (arch, column) in zip(Z,columns):
            current_data = [column]
            for value in arch:
                if type in ["OAA","RBOAA"]:
                    value *=p
                current_data.append(value)
                
            data.append(current_data)
        
        for i in range(len(Z.T)):
            names.append("Archetype {0}".format(i+1))

        df=pd.DataFrame(data,columns=names)
        df.plot(x="Attributes", y=names[1:], kind="bar",figsize=(10,10))
        plt.ylim(0.0, p+0.5)
        plt.ylabel(f"Value")
        if title == "":
            plt.title(f"Value-Distribution over All Archetypes")
        else:
            plt.title(title)
        
        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)


    def _typal_plot(self, Z, types, weighted,  save_fig, filename,title):
        plt.rcParams["figure.figsize"] = (10,10)
        fig, ax = plt.subplots()
        type_names = types.keys()
        type_names_display = list(types.keys())
        labels = [f"Archetype {i}" for i in range(len(Z.T))]
        width = 0.5
        bottoms = []
        bottoms.append([0 for i in range(len(Z.T))])
        values = []
        for label in type_names:
            label_values = []
            for archetype in Z.T:
                archetype_value = 0
                for i in types[label-1]:
                    archetype_value += archetype[i]
                if weighted in ["equal","equal_norm"]:
                    archetype_value = archetype_value / len(types[label])
                label_values.append(archetype_value)
            values.append(label_values)
        
        values_new = np.array(values)

        if weighted in ["norm","equal_norm"]:
            for i in range(len(values)):
                values_new[i] = values_new[i] / np.sum(values,0)

        for i in range(len(values)-1):
            bottoms.append([b + l for (b,l) in zip(bottoms[-1],values_new[i])])

        for i in range(len(values)):
            ax.bar(labels, values_new[i], width, bottom=bottoms[i], label=type_names_display[i])
        ax.set_ylabel('Value')
        if title == "":
            ax.set_title('Typal Composition of Archetypes')
        else:
            ax.set_title(title)
        ax.legend()

        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)
        

    def _pie_chart(self, A, indexes, attribute_indexes, archetype_dataframe,  save_fig, filename,title):

        data = []

        if attribute_indexes == []:

            for i in range(len(indexes)):
                datapoint = A.T[indexes[i]]
                if len(data) == 0:
                    data = datapoint
                else:
                    data = data + datapoint / 2

        else:
            data_subset = archetype_dataframe.copy()

            for pair in attribute_indexes:
                if pair[1] == "=":
                    data_subset = data_subset.loc[data_subset[pair[0]] == pair[2]]
                elif pair[1] == "<":
                    data_subset = data_subset.loc[data_subset[pair[0]] < pair[2]]
                elif pair[1] == ">":
                    data_subset = data_subset.loc[data_subset[pair[0]] > pair[2]]
            
            if data_subset.shape[0] < 1:
                print("\nThere are no datapoints with the value(s) given by the 'attribute_indexes' parameter.")
                return

            for i in range(data_subset.shape[0]):
                datapoint = A.T[i]
                if len(data) == 0:
                    data = datapoint
                else:
                    data = data + datapoint / 2


        labels = []
        explode = []
        for i in range(len(A.T[0])):
            labels.append("Archetype {0}".format(i+1))
            if data[i] == np.max(data):
                explode.append(0.1)
            else:
                explode.append(0.0)

        plt.pie(data,explode=tuple(explode), labels = labels, shadow=True, startangle=90, autopct='%1.1f%%')
        if title == "":
            plt.title("Pie Chart of Archetype Distribution on Given Subset of Data")
        else:
            plt.title(title)
        
        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)
        

    def _attribute_distribution(self, A, Z, indexes, columns, p, type, attribute_indexes, archetype_dataframe,  save_fig, filename, title):
        
        archetype_distribution = []

        if attribute_indexes == []:

            for i in range(len(indexes)):
                datapoint = A.T[indexes[i]]
                if len(archetype_distribution) == 0:
                    archetype_distribution = datapoint
                else:
                    archetype_distribution = archetype_distribution + datapoint / 2

        else:
            data_subset = archetype_dataframe.copy()

            for pair in attribute_indexes:
                if pair[1] == "=":
                    data_subset = data_subset.loc[data_subset[pair[0]] == pair[2]]
                elif pair[1] == "<":
                    data_subset = data_subset.loc[data_subset[pair[0]] < pair[2]]
                elif pair[1] == ">":
                    data_subset = data_subset.loc[data_subset[pair[0]] > pair[2]]
            
            if data_subset.shape[0] < 1:
                print("\nThere are no datapoints with the value {0} of attribute {1}.\n".format(attribute_indexes[1],attribute_indexes[0]))
                return

            for i in range(data_subset.shape[0]):
                datapoint = A.T[i]
                if len(archetype_distribution) == 0:
                    archetype_distribution = datapoint
                else:
                    archetype_distribution = archetype_distribution + datapoint / 2
        
        archetype_distribution = archetype_distribution/np.sum(archetype_distribution)

        attribute_distribution = []

        for a in range(len(archetype_distribution)):
            if len(attribute_distribution) == 0:
                attribute_distribution = Z.T[a]*archetype_distribution[a]
            else:
                attribute_distribution += Z.T[a]*archetype_distribution[a]

        
        plt.rcParams["figure.figsize"] = (10,10)
        if type in ["OAA","RBOAA"]:
            attribute_distribution *=p
        fig, ax = plt.subplots()
        ax.set_ylabel('Value')
        plt.xlabel('Attributes')
        if title == "":
            ax.set_title(f"Value-Distribution of Archetype")
        else:
            ax.set_title(title)
        ax.bar(np.arange(len(attribute_distribution)),attribute_distribution)
        ax.set_xticks(np.arange(len(attribute_distribution)))
        ax.set_xticklabels(labels=columns)
        plt.ylim(0, p+0.5)
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
        fig.set_size_inches(10, 10)
        
        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)
        
    
    def _circular_typal_barplot(self, type, Z, types, archetype_number,columns,p, save_fig, filename, title):

        archetype = Z.T[archetype_number-1]
        if type in ["OAA","RBOAA"]:
            archetype *=p

        type_values = []
        type_names = types.keys()
        for type in type_names:
            type_value = 0
            for attribute in types[type]:
                type_value += archetype[attribute-1]
            type_values.append(type_value/len(types[type]))
        
        
        archetype = type_values

        ANGLES = np.linspace(0.0, 2 * np.pi - 0.0, len(archetype), endpoint=False)
        width = 1/(len(archetype)/6.3)

        fig, ax = plt.subplots(figsize=(9, 12.6), subplot_kw={"projection": "polar"})
        fig.patch.set_facecolor("white")
        ax.set_facecolor("#dae5eb")

        ax.set_theta_offset(1.2 * np.pi / 2)
        ax.set_ylim(0, p)

        ax.bar(ANGLES, archetype, alpha=0.9, width=width, zorder=10)
        ax.vlines(ANGLES, 0, p, ls=(0, (4, 4)), zorder=11)

        ax.set_xticks(ANGLES)
        ax.set_xticklabels(type_names, size=12)

        if title == "":
            ax.set_title("Circular Typal Barplot of Archetype {0}".format(archetype_number),size = 25)
        else:
            ax.set_title(title,size = 25)

        if not save_fig:
            plt.show()
        else:
            plt.savefig("{0}.png".format(filename),dpi=300)

        


########## IMPORT ##########
from matplotlib.pyplot import plot
from plots_class import _plots
import pickle


########## CONVENTIONAL ARCHETYPAL ANALYSIS RESULT ##########
class _CAA_result:

    plots = _plots()
    
    def __init__(self, A, B, X, X_hat, n_iter, RSS, Z, K, p, time, columns,type, with_synthetic_data = False):
        self.A = A
        self.B = B
        self.X = X
        self.X_hat  = X_hat
        self.n_iter = len(RSS)
        self.loss = RSS
        self.Z = Z
        self.K = K
        self.p = p
        self.time = time
        self.columns = columns
        self.type = type
        self.with_synthetic_data = with_synthetic_data
        self.N = len(self.X[0,:])

    def _print(self):
        if self.type == "CAA":
            type_name = "Conventional Archetypal"
        else:
            type_name = "Two Step Archetypal"
        print("/////////////// INFORMATION ABOUT " + type_name.upper() + " ANALYSIS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\")
        print(f"▣ The " + type_name + " Analysis was computed using " + str(self.K) + " archetypes.")
        print(f"▣ The " + type_name + " Analysis was computed on " + str(len(self.X)) + " attributes.")
        print(f"▣ The " + type_name + " Analysis was computed on " + str(self.N) + " subjects.")
        print(f"▣ The " + type_name + " Analysis ran for " + str(self.n_iter) + " iterations.")
        print(f"▣ The " + type_name + " Analysis took " + str(self.time) + " seconds to complete.")
        print(f"▣ The final RSS was: {self.loss[-1]}.")

    def _plot(self,plot_type, attributes, archetype_number, types, weighted, subject_indexes, attribute_indexes, archetype_dataframe, save_figure, filename,title):
        
        if plot_type == "PCA_scatter_plot":
            self.plots._PCA_scatter_plot(self.Z,self.X_hat,self.type, save_figure, filename,title)
        elif plot_type == "attribute_scatter_plot":
            self.plots._attribute_scatter_plot(self.Z,self.X_hat,attributes,self.type,self.p, save_figure, filename,title)
        elif plot_type == "loss_plot":
            self.plots._loss_plot(self.loss,self.type,  save_figure, filename,title)
        elif plot_type == "mixture_plot":
            self.plots._mixture_plot(self.Z,self.A,self.type,  save_figure, filename,title)
        elif plot_type == "barplot":
            self.plots._barplot(self.Z,self.columns,archetype_number,self.type, self.p,  save_figure, filename,title)
        elif plot_type == "barplot_all":
            self.plots._barplot_all(self.Z,self.columns, self.type, self.p,  save_figure, filename,title)
        elif plot_type == "typal_plot":
            self.plots._typal_plot(self.Z,types,weighted,  save_figure, filename,title)
        elif plot_type == "pie_chart":
            self.plots._pie_chart(self.A, subject_indexes, attribute_indexes, archetype_dataframe,  save_figure, filename,title)
        elif plot_type == "attribute_distribution":
            self.plots._attribute_distribution(self.A,self.Z,subject_indexes,self.columns,self.p,self.type, attribute_indexes, archetype_dataframe,  save_figure, filename,title)
        elif plot_type == "circular_typal_barplot":
            self.plots._circular_typal_barplot(self.type,self.Z, types, archetype_number,self.columns,self.p, save_figure, filename, title)

    def _save(self,filename):
        if not self.with_synthetic_data:
            file = open("results/" + self.type + "_" + filename + '.obj','wb')
        else:
            file = open("synthetic_results/" + self.type + "_" + filename + '.obj','wb')
        pickle.dump(self, file)
        file.close()


########## ORDINAL ARCHETYPAL ANALYSIS RESULT ##########
class _OAA_result:

    plots = _plots()
    
    def __init__(self, A, B, X, n_iter, b, Z, X_tilde, Z_tilde, X_hat, loss, K, p, time, columns,type,sigma, with_synthetic_data = False):
        self.A = A
        self.B = B
        self.X = X
        self.n_iter = len(loss)
        self.b = b
        self.sigma = sigma
        self.X_tilde = X_tilde
        self.Z_tilde = Z_tilde
        self.X_hat = X_hat
        self.loss = loss
        self.Z = Z
        self.K = K
        self.p = p
        self.time = time
        self.columns = columns
        self.type = type
        self.with_synthetic_data = with_synthetic_data
        self.N = len(self.X[0,:])

    def _print(self):
        if self.type == "RBOAA":
            type_name = "Response Bias Ordinal Archetypal"
        else:
            type_name = "Ordinal Archetypal"
        
        print("/////////////// INFORMATION ABOUT " + type_name.upper() + " ANALYSIS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\")
        print(f"▣ The " + type_name + " Analysis was computed using " + str(self.K) + " archetypes.")
        print(f"▣ The " + type_name + " Analysis was computed on " + str(len(self.X)) + " attributes.")
        print(f"▣ The " + type_name + " Analysis was computed on " + str(self.N) + " subjects.")
        print(f"▣ The " + type_name + " Analysis ran for " + str(self.n_iter) + " iterations.")
        print(f"▣ The " + type_name + " Analysis took " + str(self.time) + " seconds to complete.")
        print(f"▣ The final loss was: {self.loss[-1]}.")
    
    def _plot(self,plot_type, attributes, archetype_number, types, weighted, subject_indexes, attribute_indexes, archetype_dataframe, save_figure, filename,title):
        
        if plot_type == "PCA_scatter_plot":
            self.plots._PCA_scatter_plot(self.Z,self.X_hat,self.type, save_figure, filename,title)
        elif plot_type == "attribute_scatter_plot":
            self.plots._attribute_scatter_plot(self.Z,self.X_hat,attributes,self.type, self.p, save_figure, filename,title)
        elif plot_type == "loss_plot":
            self.plots._loss_plot(self.loss,self.type,  save_figure, filename,title)
        elif plot_type == "mixture_plot":
            self.plots._mixture_plot(self.Z,self.A,self.type,  save_figure, filename,title)
        elif plot_type == "barplot":
            self.plots._barplot(self.Z,self.columns,archetype_number,self.type,self.p,  save_figure, filename,title)
        elif plot_type == "barplot_all":
            self.plots._barplot_all(self.Z,self.columns, self.type, self.p,  save_figure, filename,title)
        elif plot_type == "typal_plot":
            self.plots._typal_plot(self.Z,types,weighted,  save_figure, filename,title)
        elif plot_type == "pie_chart":
            self.plots._pie_chart(self.A,subject_indexes, attribute_indexes, archetype_dataframe,  save_figure, filename,title)
        elif plot_type == "attribute_distribution":
            self.plots._attribute_distribution(self.A,self.Z,subject_indexes,self.columns,self.p,self.type,attribute_indexes, archetype_dataframe,  save_figure, filename,title)
        elif plot_type == "circular_typal_barplot":
            self.plots._circular_typal_barplot(self.type,self.Z, types, archetype_number,self.columns,self.p,save_figure, filename, title)

    def _save(self,filename):
        if not self.with_synthetic_data:
            file = open("results/" + self.type + "_" + filename + '.obj','wb')
        else:
            file = open("synthetic_results/" + self.type + "_" + filename + '.obj','wb')
        pickle.dump(self, file)
        file.close()


import numpy as np
from scipy.stats import norm
import pickle

########## CLASS FOR CREATING SYNTHETIC DATA  ##########
class _synthetic_data:
    
    ########## INITIALIZATION - CREATES THE DATA ##########
    def __init__(self, N, M ,K, p, sigma, rb, a_param, b_param, sigma_std = 0):
        
        self.N = N
        self.M = M
        self.K = K
        self.p = p
        self.columns = ["SQ"+str(i) for i in range(1, M+1)]
        self.X, self.Z, self.Z_alpha, self.A, self.betas = self.X(N=N, M=M, K=K, p=p, sigma=sigma, rb=rb, a_param=a_param, b_param=b_param, sigma_std = sigma_std)
        
    ########## IF THERE IS RESPONSE BIAS IN THE DATA ##########
    def biasedBetas(self, N, p, b_param):
        b = np.array([b_param]*p)
        return np.random.dirichlet(b, size=N)
    
    ########## CONTSTRAINTS ON THE RESPONSE BIAS BETAS ##########
    def betaConstraintsBias(self, betas):
        N, J = betas.shape
        new_betas = np.empty((N,J))
    
        denoms = np.sum(betas,axis=1)
        
        for i in range(N):
            for j in range(J):
                new_betas[i,j] = np.sum(betas[i,:j+1])/denoms[i]
    
        # Return and remove the column of ones
        return new_betas[:,:-1]
    
    ########## CONSTRAINTS ON THE NON RESOPNS BIAS BETAS ##########
    def betaConstraints(self, betas):
   
       new_betas = np.empty(len(betas))
       denom = sum(betas)
       for i in range(len(new_betas)):
           new_betas[i] = np.sum(betas[:i+1]) / denom
   
       return new_betas[:-1]
   
    ########## SOFTPLUS HELPER FUNCTION ##########
    def softplus(self, sigma, sigma_std):
        if sigma_std == 0:
            return np.log(1 + np.exp(sigma))
        else:
            sigmas = []
            for n in range(self.N):
                sigmas.append(np.log(1 + np.exp(sigma + np.random.uniform(-1,1,1)*sigma_std)))
            sigmasMatrix  = np.repeat(sigmas, self.M, axis=1)
            return sigmasMatrix

    ########## HELPER FUNCTION, CALCULATES THE Z ARCEHTYPE MATRIX ##########
    def get_Z(self, N, M, K, p, rb, b_param):
        # Ensure reproducibility
        np.random.seed(42)
        
        
        
        # Check to ensure that there are no NaN's
        if b_param < 0.01:
            b_param = 0.01
        
        betas = np.array([b_param]*p)
        betas = self.betaConstraints(betas)
        
        alphas = np.empty(p)
        alphas[0] = (0 + betas[0]) / 2
        alphas[-1] = (1+ betas[-1]) / 2
        for i in range(len(betas)-1):
            alphas[i+1] = (betas[i] + betas[i+1]) / 2
        
        
        Z_ordinal = np.ceil(np.random.uniform(0, 1, size = (M,K))*p).astype(int)
        Z_alpha = alphas[Z_ordinal-1]
        
        
        if rb == True:
            betas = self.biasedBetas(N=N, p=p, b_param=b_param)
            betas = self.betaConstraintsBias(betas)
            
        return Z_ordinal, Z_alpha, betas

    ########## HELPER FUNCTION, CALCULATES THE A LINEAZR COMBINATION MATRIX ##########
    def get_A(self, N, K, a_param):
        np.random.seed(42) # set another seed :)
        
        # Constrain a_param to avoid NaN's
        if a_param < 0.01:
            a_param = 0.01
        
        alpha = np.array([a_param]*K)
        return np.random.dirichlet(alpha, size=N).transpose()
    
    ########## HELPER FUNCTION, CALCULATES THE D DENSITY MATRIX ##########
    def get_D(self, X_rec, betas, sigma, rb):
        
        M, N = X_rec.shape
        
        
        if rb == False:
        
            J = len(betas)    
            D = np.empty((J+2, M, N))
            
            for j in range(J+2):
                # Left-most tail
                if j == 0:
                    D[j] = np.ones((M,N))*(np.inf*(-1))
                # Right-most tail
                elif j == J+1:
                    D[j] = np.ones((M,N))*(np.inf)
                else:
                    D[j] = (betas[j-1] - X_rec)/(sigma.T+1e-16) ## Add softplus(sigma)
                    
        else:
            J = len(betas[0,:])
            D = np.empty((J+2, M, N))

            # D = torch.rand(len(betas[0,:])+2,M,N)
            # D[0] = torch.tensor(np.matrix(np.ones((N)) * (-np.inf)))
            # D[-1] = torch.tensor(np.matrix(np.ones((N)) * (np.inf)))
            # D[1:-1] = torch.div(torch.unsqueeze(betas.T, 2).repeat(1,1,N)-X_rec.T,torch.unsqueeze(sigma+1e-16, 1).repeat(1,N))
            
            for j in range(J+2):
                if j == 0:
                    D[j] = np.ones((M,N))*(np.inf*(-1))
                elif j == J+1:
                    D[j] = np.ones((M,N))*(np.inf)
                else:
                    D[j] = (betas[:,j-1] - X_rec)/((sigma.T+1e-16)) ## Add softplus(sigma)
                    # D[j] = torch.div((b[:,j-1] - X_hat[:, None]),sigma)[:,0,:].T
        
        # print("SHAPEEEE", D.shape)
        # min = np.min(D[(D > -np.inf) & (D < np.inf)]) 
        # max = np.max(D[(D > -np.inf) & (D < np.inf)])
        # D[J+1] = np.ones((M,N))*(max + (max-min)/J)
        # D[0] = np.ones((M,N))*(min - (max-min)/J)
        # print("min, max:", min, max)
        # print("D_0", D[0])
        return D - np.mean(D[1:-1])

    ########## HELPER FUNCTION, CALCULATES THE PROBABILITY FROM THE DENSITY MATRIX ##########
    def Probs(self, D):
        
        J, M, N = D.shape
        
        probs = np.empty((J-1, M, N)) 
        for i in range(J):
            if i != J-1:
                probs[i,:,:] = norm.cdf(D[i+1], loc=0, scale=1) - norm.cdf(D[i], loc=0, scale=1)
                
        return probs

    ########## HELPER FUNCTION, SAMPLES FROM PROBABILITY MATRIX TO GET CATEGORICAL ##########
    def toCategorical(self, probs):
        
        categories = np.arange(1, len(probs)+1)    
        J, M, N = probs.shape
        X_cat = np.empty((M,N))
        
        
        for m in range(M):
            for n in range(N):
                X_cat[m,n] = int(np.random.choice(categories, p = list(probs[:,m,n])))
                
        X_cat = X_cat.astype(int)
        return X_cat
    
    ########## CALUCLATES DATA WITH HELP OF ALL OTHER FUNCTIONS ##########
    def X(self, M, N, K, p, sigma, rb=False, a_param=1, b_param=100, sigma_std = 0):
        
        Z_ordinal, Z_alpha, betas = self.get_Z(N=N,M=M, K=K, p=p, rb=rb, b_param=b_param)
        A = self.get_A(N, K, a_param=a_param)
        X_rec = Z_alpha@A
        
        D = self.get_D(X_rec, betas, self.softplus(sigma, sigma_std), rb=rb)
        probs = self.Probs(D)
        
        X_final = self.toCategorical(probs)
        
        
        return X_final, Z_ordinal, Z_alpha, A, betas

    ########## SAVES THE DATA LOCALLY ON PC ########## 
    def _save(self,type,filename):
        file = open("synthetic_results/" + type + "_" + filename + '_metadata' + '.obj','wb')
        pickle.dump(self, file)
        file.close()



from timeit import default_timer as timer



class _loading_bar:
    
    def __init__(self, count, headline):
        self.count = count
        self.current = 0
        self.skip = 0
        self.start = timer()
        print("\n{0} in progress...".format(headline))
        
    def _update(self, loss = None):
        self.current += 1
        if self.current == self.count:
            print("\r{0}".format(f"Finished Successfully after {round(timer()-self.start,1)} seconds!                               \n"))
        elif self.skip > self.count/133:
            remaining = (self.count/self.current)*(timer()-self.start)-(timer()-self.start)
            if remaining > 60:
                remaining = f"{round(remaining/60,1)} minutes remaining"
            else:
                remaining = f"{round(remaining,1)} seconds remaining"
            if not loss == None:
                print("\r{0}".format("|{0}{1}|   {2}% finished // {3} (Loss: {4})".format("█"*int(round(self.current/self.count,1)*20),"-"*(20-int(round(self.current/self.count,1)*20)),round((self.current/self.count)*100,2),remaining, round(loss))), end = "", flush=True)
            print("\r{0}".format("|{0}{1}|   {2}% finished // {3}".format("█"*int(round(self.current/self.count,1)*20),"-"*(20-int(round(self.current/self.count,1)*20)),round((self.current/self.count)*100,2),remaining)), end = "", flush=True)
            self.skip = 0
        else:
            self.skip += 1

    def _kill(self):
        print("\r{0}".format(f"Finished Successfully after {round(timer()-self.start,1)} seconds!                               \n"))